# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OTLP Exporter"""

import enum
import logging
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from time import sleep
from typing import Any, Callable, Dict, Generic, List, Optional
from typing import Sequence as TypingSequence
from typing import Text, TypeVar

from backoff import expo
from google.rpc.error_details_pb2 import RetryInfo
from grpc import (
    ChannelCredentials,
    Compression,
    RpcError,
    StatusCode,
    insecure_channel,
    secure_channel,
    ssl_channel_credentials,
)

from opentelemetry.configuration import Configuration
from opentelemetry.proto.common.v1.common_pb2 import AnyValue, KeyValue
from opentelemetry.proto.resource.v1.resource_pb2 import Resource
from opentelemetry.sdk.resources import Resource as SDKResource

logger = logging.getLogger(__name__)
SDKDataT = TypeVar("SDKDataT")
ResourceDataT = TypeVar("ResourceDataT")
TypingResourceT = TypeVar("TypingResourceT")
ExportServiceRequestT = TypeVar("ExportServiceRequestT")
ExportResultT = TypeVar("ExportResultT")


class OTLPCompression(enum.Enum):
    gzip = "gzip"


def _translate_key_values(key: Text, value: Any) -> KeyValue:

    if isinstance(value, bool):
        any_value = AnyValue(bool_value=value)

    elif isinstance(value, str):
        any_value = AnyValue(string_value=value)

    elif isinstance(value, int):
        any_value = AnyValue(int_value=value)

    elif isinstance(value, float):
        any_value = AnyValue(double_value=value)

    elif isinstance(value, Sequence):
        any_value = AnyValue(array_value=value)

    elif isinstance(value, Mapping):
        any_value = AnyValue(kvlist_value=value)

    else:
        raise Exception(
            "Invalid type {} of value {}".format(type(value), value)
        )

    return KeyValue(key=key, value=any_value)


def _get_resource_data(
    sdk_resource_instrumentation_library_data: Dict[
        SDKResource, ResourceDataT
    ],
    resource_class: Callable[..., TypingResourceT],
    name: str,
) -> List[TypingResourceT]:

    resource_data = []

    for (
        sdk_resource,
        instrumentation_library_data,
    ) in sdk_resource_instrumentation_library_data.items():

        collector_resource = Resource()

        for key, value in sdk_resource.attributes.items():

            try:
                # pylint: disable=no-member
                collector_resource.attributes.append(
                    _translate_key_values(key, value)
                )
            except Exception as error:  # pylint: disable=broad-except
                logger.exception(error)

        resource_data.append(
            resource_class(
                **{
                    "resource": collector_resource,
                    "instrumentation_library_{}".format(name): [
                        instrumentation_library_data
                    ],
                }
            )
        )

    return resource_data


def _load_credential_from_file(filepath) -> ChannelCredentials:
    try:
        with open(filepath, "rb") as f:
            credential = f.read()
            return ssl_channel_credentials(credential)
    except FileNotFoundError:
        logger.exception("Failed to read credential file")
        return None


# pylint: disable=no-member
class OTLPExporterMixin(
    ABC, Generic[SDKDataT, ExportServiceRequestT, ExportResultT]
):
    """OTLP span/metric exporter

    Args:
        endpoint: OpenTelemetry Collector receiver endpoint
        insecure: Connection type
        credentials: ChannelCredentials object for server authentication
        headers: Headers to send when exporting
        compression: Compression algorithm to be used in channel
        timeout: Backend request timeout in seconds
    """

    def __init__(
        self,
        endpoint: Optional[str] = None,
        insecure: Optional[bool] = None,
        credentials: Optional[ChannelCredentials] = None,
        headers: Optional[str] = None,
        timeout: Optional[int] = None,
        compression: str = None,
    ):
        super().__init__()

        endpoint = (
            endpoint
            or Configuration().EXPORTER_OTLP_ENDPOINT
            or "localhost:55680"
        )

        if insecure is None:
            insecure = Configuration().EXPORTER_OTLP_INSECURE
        if insecure is None:
            insecure = False

        self._headers = headers or Configuration().EXPORTER_OTLP_HEADERS
        self._timeout = (
            timeout
            or Configuration().EXPORTER_OTLP_TIMEOUT
            or 10  # default: 10 seconds
        )
        self._collector_span_kwargs = None

        if compression is None:
            compression_algorithm = Compression.NoCompression
        elif (
            compression in OTLPCompression._value2member_map_
            and OTLPCompression(compression) is OTLPCompression.gzip
        ):
            compression_algorithm = Compression.Gzip
        else:
            compression_str = Configuration().EXPORTER_OTLP_INSECURE or None
            if compression_str is None:
                compression_algorithm = Compression.NoCompression
            elif (
                compression_str in OTLPCompression._value2member_map_
                and OTLPCompression(compression_str) is OTLPCompression.gzip
            ):
                compression_algorithm = Compression.Gzip
            else:
                raise ValueError(
                    "OTEL_EXPORTER_OTLP_COMPRESSION environment variable does not match gzip."
                )

        if insecure:
            self._client = self._stub(
                insecure_channel(endpoint, compression=compression_algorithm)
            )
            return

        # secure mode
        if (
            credentials is None
            and Configuration().EXPORTER_OTLP_CERTIFICATE is None
        ):
            raise ValueError("No credentials set in secure mode.")

        credentials = credentials or _load_credential_from_file(
            Configuration().EXPORTER_OTLP_CERTIFICATE
        )
        self._client = self._stub(
            secure_channel(
                endpoint, credentials, compression=compression_algorithm
            )
        )

    @abstractmethod
    def _translate_data(
        self, data: TypingSequence[SDKDataT]
    ) -> ExportServiceRequestT:
        pass

    def _export(self, data: TypingSequence[SDKDataT]) -> ExportResultT:
        # expo returns a generator that yields delay values which grow
        # exponentially. Once delay is greater than max_value, the yielded
        # value will remain constant.
        # max_value is set to 900 (900 seconds is 15 minutes) to use the same
        # value as used in the Go implementation.

        max_value = 900

        for delay in expo(max_value=max_value):

            if delay == max_value:
                return self._result.FAILURE

            try:
                self._client.Export(
                    request=self._translate_data(data),
                    metadata=self._headers,
                    timeout=self._timeout,
                )

                return self._result.SUCCESS

            except RpcError as error:

                if error.code() in [
                    StatusCode.CANCELLED,
                    StatusCode.DEADLINE_EXCEEDED,
                    StatusCode.PERMISSION_DENIED,
                    StatusCode.UNAUTHENTICATED,
                    StatusCode.RESOURCE_EXHAUSTED,
                    StatusCode.ABORTED,
                    StatusCode.OUT_OF_RANGE,
                    StatusCode.UNAVAILABLE,
                    StatusCode.DATA_LOSS,
                ]:

                    retry_info_bin = dict(error.trailing_metadata()).get(
                        "google.rpc.retryinfo-bin"
                    )
                    if retry_info_bin is not None:
                        retry_info = RetryInfo()
                        retry_info.ParseFromString(retry_info_bin)
                        delay = (
                            retry_info.retry_delay.seconds
                            + retry_info.retry_delay.nanos / 1.0e9
                        )

                    logger.debug(
                        "Waiting %ss before retrying export of span", delay
                    )
                    sleep(delay)
                    continue

                if error.code() == StatusCode.OK:
                    return self._result.SUCCESS

                return self._result.FAILURE

        return self._result.FAILURE

    def shutdown(self) -> None:
        pass



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import OrderedDict
from unittest import TestCase
from unittest.mock import patch

from grpc import ChannelCredentials

from opentelemetry.configuration import Configuration
from opentelemetry.exporter.otlp.metrics_exporter import OTLPMetricsExporter
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2 import (
    ExportMetricsServiceRequest,
)
from opentelemetry.proto.common.v1.common_pb2 import (
    AnyValue,
    InstrumentationLibrary,
    KeyValue,
    StringKeyValue,
)
from opentelemetry.proto.metrics.v1.metrics_pb2 import (
    AggregationTemporality,
    InstrumentationLibraryMetrics,
    IntDataPoint,
    IntSum,
)
from opentelemetry.proto.metrics.v1.metrics_pb2 import Metric as OTLPMetric
from opentelemetry.proto.metrics.v1.metrics_pb2 import ResourceMetrics
from opentelemetry.proto.resource.v1.resource_pb2 import (
    Resource as OTLPResource,
)
from opentelemetry.sdk.metrics import Counter, MeterProvider
from opentelemetry.sdk.metrics.export import ExportRecord
from opentelemetry.sdk.metrics.export.aggregate import SumAggregator
from opentelemetry.sdk.resources import Resource as SDKResource

THIS_DIR = os.path.dirname(__file__)


class TestOTLPMetricExporter(TestCase):
    @patch("opentelemetry.sdk.metrics.export.aggregate.time_ns")
    def setUp(self, mock_time_ns):  # pylint: disable=arguments-differ
        mock_time_ns.configure_mock(**{"return_value": 1})
        self.exporter = OTLPMetricsExporter(insecure=True)
        resource = SDKResource(OrderedDict([("a", 1), ("b", False)]))

        self.counter_export_record = ExportRecord(
            Counter(
                "c",
                "d",
                "e",
                int,
                MeterProvider(resource=resource,).get_meter("name", "version"),
                ("f",),
            ),
            [("g", "h")],
            SumAggregator(),
            resource,
        )

        Configuration._reset()  # pylint: disable=protected-access

    def tearDown(self):
        Configuration._reset()  # pylint: disable=protected-access

    @patch.dict(
        "os.environ",
        {
            "OTEL_EXPORTER_OTLP_METRIC_ENDPOINT": "collector:55680",
            "OTEL_EXPORTER_OTLP_METRIC_CERTIFICATE": THIS_DIR
            + "/fixtures/test.cert",
            "OTEL_EXPORTER_OTLP_METRIC_HEADERS": "key1:value1;key2:value2",
            "OTEL_EXPORTER_OTLP_METRIC_TIMEOUT": "10",
        },
    )
    @patch("opentelemetry.exporter.otlp.exporter.OTLPExporterMixin.__init__")
    def test_env_variables(self, mock_exporter_mixin):
        OTLPMetricsExporter()

        self.assertTrue(len(mock_exporter_mixin.call_args_list) == 1)
        _, kwargs = mock_exporter_mixin.call_args_list[0]

        self.assertEqual(kwargs["endpoint"], "collector:55680")
        self.assertEqual(kwargs["headers"], "key1:value1;key2:value2")
        self.assertEqual(kwargs["timeout"], 10)
        self.assertIsNotNone(kwargs["credentials"])
        self.assertIsInstance(kwargs["credentials"], ChannelCredentials)

    def test_no_credentials_error(self):
        with self.assertRaises(ValueError):
            OTLPMetricsExporter()

    def test_translate_metrics(self):
        # pylint: disable=no-member

        self.counter_export_record.aggregator.checkpoint = 1
        self.counter_export_record.aggregator.initial_checkpoint_timestamp = 1
        self.counter_export_record.aggregator.last_update_timestamp = 1

        expected = ExportMetricsServiceRequest(
            resource_metrics=[
                ResourceMetrics(
                    resource=OTLPResource(
                        attributes=[
                            KeyValue(key="a", value=AnyValue(int_value=1)),
                            KeyValue(
                                key="b", value=AnyValue(bool_value=False)
                            ),
                        ]
                    ),
                    instrumentation_library_metrics=[
                        InstrumentationLibraryMetrics(
                            instrumentation_library=InstrumentationLibrary(
                                name="name", version="version",
                            ),
                            metrics=[
                                OTLPMetric(
                                    name="c",
                                    description="d",
                                    unit="e",
                                    int_sum=IntSum(
                                        data_points=[
                                            IntDataPoint(
                                                labels=[
                                                    StringKeyValue(
                                                        key="g", value="h"
                                                    )
                                                ],
                                                value=1,
                                                time_unix_nano=1,
                                                start_time_unix_nano=1,
                                            )
                                        ],
                                        aggregation_temporality=(
                                            AggregationTemporality.AGGREGATION_TEMPORALITY_CUMULATIVE
                                        ),
                                        is_monotonic=True,
                                    ),
                                )
                            ],
                        )
                    ],
                )
            ]
        )

        # pylint: disable=protected-access
        actual = self.exporter._translate_data([self.counter_export_record])

        self.assertEqual(expected, actual)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OTLP Metrics Exporter"""

import logging
from typing import List, Optional, Sequence, Type, TypeVar

from grpc import ChannelCredentials

from opentelemetry.configuration import Configuration
from opentelemetry.exporter.otlp.exporter import (
    OTLPExporterMixin,
    _get_resource_data,
    _load_credential_from_file,
)
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2 import (
    ExportMetricsServiceRequest,
)
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2_grpc import (
    MetricsServiceStub,
)
from opentelemetry.proto.common.v1.common_pb2 import StringKeyValue
from opentelemetry.proto.metrics.v1.metrics_pb2 import (
    AggregationTemporality,
    DoubleDataPoint,
    DoubleGauge,
    DoubleSum,
    InstrumentationLibraryMetrics,
    IntDataPoint,
    IntGauge,
    IntSum,
)
from opentelemetry.proto.metrics.v1.metrics_pb2 import Metric as OTLPMetric
from opentelemetry.proto.metrics.v1.metrics_pb2 import ResourceMetrics
from opentelemetry.sdk.metrics import (
    Counter,
    SumObserver,
    UpDownCounter,
    UpDownSumObserver,
    ValueObserver,
    ValueRecorder,
)
from opentelemetry.sdk.metrics.export import (
    ExportRecord,
    MetricsExporter,
    MetricsExportResult,
)
from opentelemetry.sdk.metrics.export.aggregate import (
    HistogramAggregator,
    LastValueAggregator,
    MinMaxSumCountAggregator,
    SumAggregator,
    ValueObserverAggregator,
)

logger = logging.getLogger(__name__)
DataPointT = TypeVar("DataPointT", IntDataPoint, DoubleDataPoint)


def _get_data_points(
    export_record: ExportRecord,
    data_point_class: Type[DataPointT],
    aggregation_temporality: int,
) -> List[DataPointT]:

    if isinstance(export_record.aggregator, SumAggregator):
        value = export_record.aggregator.checkpoint

    elif isinstance(export_record.aggregator, MinMaxSumCountAggregator):
        # FIXME: How are values to be interpreted from this aggregator?
        raise Exception("MinMaxSumCount aggregator data not supported")

    elif isinstance(export_record.aggregator, HistogramAggregator):
        # FIXME: How are values to be interpreted from this aggregator?
        raise Exception("Histogram aggregator data not supported")

    elif isinstance(export_record.aggregator, LastValueAggregator):
        value = export_record.aggregator.checkpoint

    elif isinstance(export_record.aggregator, ValueObserverAggregator):
        value = export_record.aggregator.checkpoint.last

    if aggregation_temporality == (
        AggregationTemporality.AGGREGATION_TEMPORALITY_CUMULATIVE
    ):
        start_time_unix_nano = export_record.aggregator.first_timestamp
    else:
        start_time_unix_nano = (
            export_record.aggregator.initial_checkpoint_timestamp
        )

    return [
        data_point_class(
            labels=[
                StringKeyValue(key=str(label_key), value=str(label_value))
                for label_key, label_value in export_record.labels
            ],
            value=value,
            start_time_unix_nano=start_time_unix_nano,
            time_unix_nano=(export_record.aggregator.last_update_timestamp),
        )
    ]


class OTLPMetricsExporter(
    MetricsExporter,
    OTLPExporterMixin[
        ExportRecord, ExportMetricsServiceRequest, MetricsExportResult
    ],
):
    # pylint: disable=unsubscriptable-object
    """OTLP metrics exporter

    Args:
        endpoint: OpenTelemetry Collector receiver endpoint
        insecure: Connection type
        credentials: Credentials object for server authentication
        headers: Headers to send when exporting
        timeout: Backend request timeout in seconds
    """

    _stub = MetricsServiceStub
    _result = MetricsExportResult

    def __init__(
        self,
        endpoint: Optional[str] = None,
        insecure: Optional[bool] = None,
        credentials: Optional[ChannelCredentials] = None,
        headers: Optional[str] = None,
        timeout: Optional[int] = None,
    ):
        if insecure is None:
            insecure = Configuration().EXPORTER_OTLP_METRIC_INSECURE

        if (
            not insecure
            and Configuration().EXPORTER_OTLP_METRIC_CERTIFICATE is not None
        ):
            credentials = credentials or _load_credential_from_file(
                Configuration().EXPORTER_OTLP_METRIC_CERTIFICATE
            )

        super().__init__(
            **{
                "endpoint": endpoint
                or Configuration().EXPORTER_OTLP_METRIC_ENDPOINT,
                "insecure": insecure,
                "credentials": credentials,
                "headers": headers
                or Configuration().EXPORTER_OTLP_METRIC_HEADERS,
                "timeout": timeout
                or Configuration().EXPORTER_OTLP_METRIC_TIMEOUT,
            }
        )

    # pylint: disable=no-self-use
    def _translate_data(
        self, export_records: Sequence[ExportRecord]
    ) -> ExportMetricsServiceRequest:
        # pylint: disable=too-many-locals,no-member
        # pylint: disable=attribute-defined-outside-init

        sdk_resource_instrumentation_library_metrics = {}

        # The criteria to decide how to translate export_records is based on this table
        # taken directly from OpenTelemetry Proto v0.5.0:

        # TODO: Update table after the decision on:
        # https://github.com/open-telemetry/opentelemetry-specification/issues/731.
        # By default, metrics recording using the OpenTelemetry API are exported as
        # (the table does not include MeasurementValueType to avoid extra rows):
        #
        #   Instrument         Type
        #   ----------------------------------------------
        #   Counter            Sum(aggregation_temporality=delta;is_monotonic=true)
        #   UpDownCounter      Sum(aggregation_temporality=delta;is_monotonic=false)
        #   ValueRecorder      TBD
        #   SumObserver        Sum(aggregation_temporality=cumulative;is_monotonic=true)
        #   UpDownSumObserver  Sum(aggregation_temporality=cumulative;is_monotonic=false)
        #   ValueObserver      Gauge()
        for export_record in export_records:

            if export_record.resource not in (
                sdk_resource_instrumentation_library_metrics.keys()
            ):
                sdk_resource_instrumentation_library_metrics[
                    export_record.resource
                ] = InstrumentationLibraryMetrics()

            type_class = {
                int: {
                    "sum": {"class": IntSum, "argument": "int_sum"},
                    "gauge": {"class": IntGauge, "argument": "int_gauge"},
                    "data_point_class": IntDataPoint,
                },
                float: {
                    "sum": {"class": DoubleSum, "argument": "double_sum"},
                    "gauge": {
                        "class": DoubleGauge,
                        "argument": "double_gauge",
                    },
                    "data_point_class": DoubleDataPoint,
                },
            }

            value_type = export_record.instrument.value_type

            sum_class = type_class[value_type]["sum"]["class"]
            gauge_class = type_class[value_type]["gauge"]["class"]
            data_point_class = type_class[value_type]["data_point_class"]

            if isinstance(export_record.instrument, Counter):

                aggregation_temporality = (
                    AggregationTemporality.AGGREGATION_TEMPORALITY_CUMULATIVE
                )

                otlp_metric_data = sum_class(
                    data_points=_get_data_points(
                        export_record,
                        data_point_class,
                        aggregation_temporality,
                    ),
                    aggregation_temporality=aggregation_temporality,
                    is_monotonic=True,
                )
                argument = type_class[value_type]["sum"]["argument"]

            elif isinstance(export_record.instrument, UpDownCounter):

                aggregation_temporality = (
                    AggregationTemporality.AGGREGATION_TEMPORALITY_CUMULATIVE
                )

                otlp_metric_data = sum_class(
                    data_points=_get_data_points(
                        export_record,
                        data_point_class,
                        aggregation_temporality,
                    ),
                    aggregation_temporality=aggregation_temporality,
                    is_monotonic=False,
                )
                argument = type_class[value_type]["sum"]["argument"]

            elif isinstance(export_record.instrument, (ValueRecorder)):
                logger.warning("Skipping exporting of ValueRecorder metric")
                continue

            elif isinstance(export_record.instrument, SumObserver):

                aggregation_temporality = (
                    AggregationTemporality.AGGREGATION_TEMPORALITY_CUMULATIVE
                )

                otlp_metric_data = sum_class(
                    data_points=_get_data_points(
                        export_record,
                        data_point_class,
                        aggregation_temporality,
                    ),
                    aggregation_temporality=aggregation_temporality,
                    is_monotonic=True,
                )
                argument = type_class[value_type]["sum"]["argument"]

            elif isinstance(export_record.instrument, UpDownSumObserver):

                aggregation_temporality = (
                    AggregationTemporality.AGGREGATION_TEMPORALITY_CUMULATIVE
                )

                otlp_metric_data = sum_class(
                    data_points=_get_data_points(
                        export_record,
                        data_point_class,
                        aggregation_temporality,
                    ),
                    aggregation_temporality=aggregation_temporality,
                    is_monotonic=False,
                )
                argument = type_class[value_type]["sum"]["argument"]

            elif isinstance(export_record.instrument, (ValueObserver)):
                otlp_metric_data = gauge_class(
                    data_points=_get_data_points(
                        export_record,
                        data_point_class,
                        AggregationTemporality.AGGREGATION_TEMPORALITY_DELTA,
                    )
                )
                argument = type_class[value_type]["gauge"]["argument"]

            instrumentation_library_metrics = sdk_resource_instrumentation_library_metrics[
                export_record.resource
            ]

            instrumentation_library_metrics.metrics.append(
                OTLPMetric(
                    **{
                        "name": export_record.instrument.name,
                        "description": (export_record.instrument.description),
                        "unit": export_record.instrument.unit,
                        argument: otlp_metric_data,
                    }
                )
            )

            instrumentation_library_metrics.instrumentation_library.name = (
                export_record.instrument.meter.instrumentation_info.name
            )

            version = (
                export_record.instrument.meter.instrumentation_info.version
            )

            if version:
                (
                    instrumentation_library_metrics.instrumentation_library.version
                ) = version

        return ExportMetricsServiceRequest(
            resource_metrics=_get_resource_data(
                sdk_resource_instrumentation_library_metrics,
                ResourceMetrics,
                "metrics",
            )
        )

    def export(self, metrics: Sequence[ExportRecord]) -> MetricsExportResult:
        # pylint: disable=arguments-differ
        return self._export(metrics)
